
local TradeSkills
local TradeSkillNames

local SVData = {}

------------------[[ Map Pin Functionality ]]-------------------
local MapPinManager = ZO_WorldMapPins:Subclass()
local CompassPinManager = ZO_ControlPool:Subclass()

-- Create the compass manager pin using the template in the xml file
function CompassPinManager:New( ... )
	local CompassManager = ZO_ControlPool.New(self, "XRGACompassPin", ZO_CompassContainer, XRGA.AddonInfo.Prefix.."Pin")
	CompassManager:Initialize( ... )	
	return CompassManager
end

-- Initialise the new compass manager and link it to the map pin manager
function CompassPinManager:Initialize(MapPinManager)
	self.Data = {}
	self.Layouts = {}
	self.Filters = {}
	self.MapPinManager = MapPinManager
end

-- Create the requested pinType and set the layout
function CompassPinManager:CreatePinType( pinType, pinLayout )
	if not self.Data and not self.Layouts and not self.MapPinManager then return end
	pinLayout.size = 20
	self.Layouts[ pinType ] = pinLayout
	self.Data[ pinType ] = {}
end

-- Create the specific pin at the specified location
function CompassPinManager:CreatePin( pinType, pinX, pinY, pinZ, pinRadius)
	if not self.Data and not self.Layouts and not self.MapPinManager then return end
	local pin, pinKey = self:AcquireObject()

	-- Store a name for this compass pin ( to hopefully go above the compass chart )
	table.insert( self.Data[ pinType ], { Key = pinKey, Name = nil } )

	pin.x = pinX
	pin.y = pinY
	pin.z = pinZ
	pin.radius = pinRadius
	pin.type = pinType

	local layout = self.Layouts[ pinType ]
	local texture = pin:GetNamedChild( "Background" )
	texture:SetTexture( layout.texture )

	-- Return the key to this pin so that the map manager can update the name with the node name
	return pinKey
end

-- Remove the selected pin type
function CompassPinManager:RemovePins( pinType )
	if not self.Data then return end
	if not pinType then
		self:ReleaseAllObjects()
		for pinType, pinData in pairs( self.Data ) do
			self.Data[ pinType ] = {}
		end
	else
		for pinID, pinKey in pairs( self.Data[ pinType ] ) do
			self:ReleaseObject( pinKey )
		end
		self.Data[ pinType ] = {}
	end
end

-- Add filters
function CompassPinManager:AddFilter(...)
	local pinType,pinID,enable = ...
	self.Filters[pinType] = enable
end

-- Triangle OAB - O = Origin x,y, A = Destination x,y, B = Difference x,y
function CompassPinManager:Update( cx, cy, heading )

	if not self then return end
	if not self.Data then return end

	for pinID, pinData in pairs(self.Data) do
		for i,v in pairs(pinData) do
			if pinData then
				local pin = self:GetExistingObject(v.Key) 
				if pin then

					-- Time for some Pythagorus Theory
					-- OB = adjacent     AB = opposite      OA = hypotenuse
					-- AB = O.y - A.y    OB = O.x - A.x     OA = sqr root of sum of the squares of OB and AB					
					local xDif = cx - pin.x  -- OB Length
					local yDif = cy - pin.y  -- AB Length
					local distance = math.sqrt((xDif * xDif) + (yDif * yDif))   -- OA Length

					-- If distance is less than 0.05 then we are pretty close
					if self.Filters[pin.type] and distance < 0.05 then
						local angle = -math.atan2(xDif,yDif)
						angle = angle + heading

						if angle > math.pi then
							angle = angle - 2 * math.pi
						elseif angle < -math.pi then
							angle = angle + 2 * math.pi
						end
						local normalizedAngle = 2 * angle / (math.pi * 1.1)
						pin:ClearAnchors()
						pin:SetAnchor(CENTER,pin:GetParent(),CENTER,pin:GetParent():GetWidth() * normalizedAngle,0)
						pin:SetHidden(math.abs(normalizedAngle) > 1.0)
						if math.abs(normalizedAngle) > 0.75 then
							pin:SetDimensions( 15 * math.abs(normalizedAngle), 15 * math.abs(normalizedAngle) )
							pin:SetAlpha( -4 * (math.abs(normalizedAngle) - 0.5) )
						elseif math.abs(normalizedAngle) > 0.50 then
							pin:SetAlpha( -6 * (math.abs(normalizedAngle) - 0.5) )
							pin:SetDimensions( 20 * math.abs(normalizedAngle), 20 * math.abs(normalizedAngle) )
						elseif math.abs(normalizedAngle) > 0.25 then
							pin:SetDimensions( 25 * math.abs(normalizedAngle), 25 * math.abs(normalizedAngle) )
							pin:SetAlpha( -8 * (math.abs(normalizedAngle) - 0.5) )
						else
							pin:SetDimensions( 30 , 30  )
							pin:SetAlpha( 1 )

							-- If we are practically looking at it then make the icon bigger
							if math.abs(normalizedAngle) < 0.05 then								
								pin:SetDimensions( 50 , 50  )
							end
						end
					else
						pin:SetHidden(true)
					end
				end
			end
		end
	end
end

-- Create the map pin manager
function MapPinManager:New()
	local pinObject = ZO_WorldMapPins.New(self)
	pinObject:Initialise()
	return pinObject
end

-- Initialise the various tables
function MapPinManager:Initialise()
	self.Info = {}
	self.Data = {}
	self.Layouts = {}
	self.LayoutCallback = {}
	self.Filters = {}
	self.CompassManager = CompassPinManager:New(self)
end

-- Create the pin
function MapPinManager:CreatePin(...)
	local pinType, pinKey, pinZone, pinX, pinY, pinZ, pinRadius = ...
	self.Data[pinType] = self.Data[pinType] or {}
	self.Data[pinType][pinZone] = self.Data[pinType][pinZone] or {}
	table.insert(self.Data[pinType][pinZone],{ 
		["key"] = pinKey, 
		["zone"] = pinZone, 
		["x"] = pinX, 
		["y"] = pinY, 
		["z"] = pinZ,
		["radius"] = pinRadius 
	})	

	-- Get the key to the appropriate compass entry so that we can enhance it
	local compassKey = self.CompassManager:CreatePin(pinType, pinX, pinY, pinRadius)	

	-- Pass the node name to the compassManager so that it can be used
	for i,v in pairs(self.CompassManager.Data[pinType]) do
		if v.Key == compassKey then
			v.Name = pinKey.Node
			break
		end
	end
end

-- Add the custom pin Type
function MapPinManager:AddCustomPin(...)
	local pinType, layoutCallback, resizeCallback, pinLayout, pinTooltipCreator, pinID, enablePin, pinName, pinIcon = ...
	self.LayoutCallback[pinType] = layoutCallback
	self.Info[pinType] = {
			ID = pinID,
			Enabled = enablePin,
			Name = pinName,
			Icon = pinIcon,
	}
	self.Layouts[pinType] = {
		level = pinLayout.level,
		texture = pinLayout.texture,
		size = pinLayout.size,
		maxDistance = pinLayout.maxDistance,
	}
	ZO_WorldMap_AddCustomPin(pinType, layoutCallback, resizeCallback, pinLayout, pinTooltipCreator)
	self:SetCustomPinEnabled( pinType, enablePin )
	self.CompassManager:CreatePinType( pinType, pinLayout )
	self:RefreshCustomPins(pinType)
end

-- Remove the respective pins
function MapPinManager:RemovePins(...)
	local pinType = ...
	self.Data[pinType] = {}
	self.CompassManager:RemovePins(pinType)
	self:RefreshCustomPins(pinType)
end

-- Add the filters
function MapPinManager:AddFilter(...)
	local pinType,pinID,enable = ...
	self.Filters[pinType] = enable
	self.CompassManager:AddFilter(...)
end

-- Refresh the pins of the selected type
function MapPinManager:RefreshCustomPins(pinType)
	ZO_WorldMap_RefreshCustomPinsOfType(_G[pinType])
end

-- Enable the pins of the selected type
function MapPinManager:SetCustomPinEnabled(pinType,bool)
	ZO_WorldMap_SetCustomPinEnabled(_G[pinType],bool)
end

function MapPinManager:IsCustomPinEnabled(pinType)
	return ZO_WorldMap_IsCustomPinEnabled(_G[pinType])
end

function MapPinManager:IsPinGroupShown(pinType)
	return ZO_WorldMap_IsPinGroupShown(_G[pinType])	
end

-- Create the usable value for the pin management
local XRGAPins = MapPinManager:New()

function XRGA:GetMapPinManager()
	return XRGAPins
end

-- Monitor the compass pin updates
local lastCompassUpdate
local compassTimer = 0.05
function XrysGatherer_CompassUpdate(self)

	if not XrysGathererWindow or not XrysGathererWindow.Coordinates then return end

	if not XRGAPins then return end

	lastCompassUpdate = lastCompassUpdate or 0

	local timeNow = GetTimeStamp()
	local timeDiff = GetDiffBetweenTimeStamps(timeNow,lastCompassUpdate)

	if timeDiff < compassTimer then return end

	-- Get the Players current location in the world
 	local mapName,xPos,yPos,zPos,heading = XRGA:GetPlayerPosition(false)
	XrysGathererWindow.Coordinates:SetText(string.format("(%0.3f,%0.3f,%0.03f)",xPos,yPos,zPos))

	if not heading then
		return
	end
	if heading > math.pi then
		heading = heading - 2 * math.pi
	end

	XRGAPins.CompassManager:Update( xPos, yPos, heading )
	lastCompassUpdate = GetTimeStamp()
end

-- Initialise the Pin Types
local function InitPinTypes()	
	for i,v in ipairs(TradeSkills) do
		local pinType = XRGA.AddonInfo.Prefix.."_"..v.Name 
		local pinID = i

		-- This is called when the map is redrawn
		local layoutCallback = function(pinManager) 

			-- Too soon as the tradeskills table doesn't exist yet
			if not TradeSkills then return end

			-- Get the Map name
		 	local pinZone = GetMapName() 

		 	-- Update the filter settings in case the options were changed since logging in
		 	XRGAPins:AddFilter(pinType,pinID,SVData["Options"].Filters[pinID])

		 	-- Create the pins for the current map
			local harvestData 
			if XRGAPins and XRGAPins.Data and XRGAPins.Data[pinType] and XRGAPins.Data[pinType][pinZone] then 
				harvestData = XRGAPins.Data[pinType][pinZone] 
			end
			if harvestData then
				for i,v in pairs(harvestData) do
					if XRGAPins.Filters[pinType] then
						pinManager:CreatePin(_G[pinType], v.key, v.x, v.y, v.radius)
					end
				end
			end

		end

		-- This is call when the map is resized - 
		local resizeCallback = nil --function(width,height) XMPins:RefreshCustomPins(pinType) end		

		-- This defines the base requirements for a map and compass pin
		local pinLayoutData = { level = 20, texture = TradeSkills[pinID].Icon, size = 10, maxDistance = 0.5 }

		-- A simple tooltip display
		local pinTooltipCreator = { creator = function(pin) 
				InformationTooltip:AddLine(pin.m_PinTag.Node) 
			end, 
			tooltip = InformationTooltip }

		-- Add the custom pin and enable it
		XRGAPins:AddCustomPin(pinType,layoutCallback,resizeCallback,pinLayoutData,pinTooltipCreator,pinID,true,TradeSkills[pinID].Name,TradeSkills[pinID].Icon) 

		-- Add a filter based on users options
		XRGAPins:AddFilter(pinType,pinID,SVData["Options"].Filters[pinID])
	end	


	-- Build item and node detail tables for easy access
	local itemInfo, nodeInfo
    for default,sv in pairs(XrysGatherer_SavedVariables) do 
		for account,accountv in pairs(sv) do
			for accountWide,acWideV in pairs(accountv) do
				for i,v in pairs(acWideV) do
					if i == "ItemData" then itemInfo = v end
					if i == "NodeData" then nodeInfo = v end
				end
			end
		end
	end

	-- Read through the node and item details to get the information to build the pin details
	for nodeTimeStamp,nodeData in pairs(nodeInfo) do
		if nodeTimeStamp ~= "version" and nodeData and type(nodeData.Items) == "table" then
			local nodeZone = nodeData.MapName
			local nodeName = nodeData.Name
			local nodeX = nodeData.X
			local nodeY = nodeData.Y
			for itemIndex,itemData in ipairs(nodeData.Items) do
				local nodeItem = itemData.Item
				if itemInfo[nodeItem] then
					local itemTrade = itemInfo[nodeItem].TradeSkill
					local pinID = itemTrade
					if pinID > 0 then
						local pinType = XRGA.AddonInfo.Prefix.."_"..TradeSkills[pinID].Name
						local pinKey = { 
							["ID"] = pinID,
							["Item"] = nodeItem, 
							["Map"] = nodeZone, 
							["X"] = nodeX, 
							["Y"] = nodeY, 
							["Node"] = nodeName, 
							["Icon"] = itemInfo[nodeItem].Icon, 
							["Type"] = itemInfo[nodeItem].Type, 
							["TradeSkill"] = itemTrade, 
							["Link"] = itemInfo[nodeItem].Link,
						}
						XRGA:CreatePin(pinType,pinKey,nodeZone,nodeX,nodeY,nil)
					else
						CHAT_SYSTEM:AddMessage("No Item Data / No Pin : "..tostring(nodeTimeStamp).." "..tostring(nodeName).." "..tostring(nodeZone).." "..tostring(nodeItem))	
					end
				end
			end
		end
	end

	-- Make sure that we start with the map at the players current location and forces a redraw of the map
	SetMapToPlayerLocation()
end

-- Create the actual pin ( used in this file and NodeInteract )
function XRGA:CreatePin(pinType,pinKey,mapName,pinX,pinY,pinRadius)
	XRGAPins:CreatePin(pinType,pinKey,mapName,pinX,pinY,pinRadius)
	XRGAPins:RefreshCustomPins(pinType)
end

--------------------------------------------------------------

-- Fill the Trade Skill Tables and initialise the pin types when the player is first activated
local function PlayerActivated(eventID)
	EVENT_MANAGER:UnregisterForEvent(XRGA.AddonInfo.Prefix.."_HarvestPins",eventID)
	TradeSkills,TradeSkillNames = XRGA:InitTradeSkillInfo()
	InitPinTypes()	
end

-- Make sure we grab the Saved Variables table as needed when the addon is first loaded
local function AddOnLoaded(eventID,addonName)
	if addonName == XRGA.AddonInfo.Name then
		EVENT_MANAGER:UnregisterForEvent(XRGA.AddonInfo.Prefix.."_HarvestPins",eventID)
		SVData = XRGA:InitialiseSV()
		EVENT_MANAGER:RegisterForEvent(XRGA.AddonInfo.Prefix.."_HarvestPins",EVENT_PLAYER_ACTIVATED,PlayerActivated)
	end
end

EVENT_MANAGER:RegisterForEvent(XRGA.AddonInfo.Prefix.."_HarvestPins",EVENT_ADD_ON_LOADED,AddOnLoaded)
